/*
    FEARLUS model0-7-1: CautiousImitativeOptimumStrategy.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1, an agent-based model of land use
    change.

    FEARLUS model0-7-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the CautiousImitativeOptimumStrategy object. 

*/

#import "CautiousImitativeOptimumStrategy.h"
#import "Number.h"
#import "AssocArray.h"
#import "SelectUseBucket.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "LandManager.h"
#import "Environment.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "BitString.h"
#import "MiscFunc.h"
#import "Interval.h"

@implementation CautiousImitativeOptimumStrategy
/*

create:withManager:andParameters:

Create the strategy, setting the manager and parameters.

*/

+(id <ImitativeStrategy, HistoricalStrategy>)create: aZone
	   withManager: (LandManager *)lmgr
	 andParameters: (Parameter *)p {
  CautiousImitativeOptimumStrategy *obj;

  obj = [super create: aZone];
  obj->lm = lmgr;
  obj->parameter = p;
  return obj;
}

/*

decideLandUseForParcel:

Use a cautious imitative optimum strategy to choose the land use, making use
of variance information. Neighbourhood weighting is ignored, and so is
neighbourhood noise..

*/

-(LandUse *)decideLandUseForParcel: (LandParcel *)lp {
  SelectUseBucket *bucket;
  id parcelIndex, tempZone;
  LandParcel *myParcel, *neighbouringParcel;
  LandManager *neighbour;
  LandUse *lu;
  AssocArray *vararr;
  int i;
  unsigned yield;
  unsigned match;
  BitString *bs;
  Interval *getExpectedYieldVar(unsigned yield, unsigned antimatch,
				unsigned lubits, Interval *var) {
				// N.B. var is call by reference
    Interval *eyield = [Interval create];
    Interval *eyield2 = [Interval create];
    Interval *total = [Interval create];
    int i;

    [total set: [Interval
		  castULL: [MiscFunc combineN: (int)lubits r: (int)yield]]];
    [eyield set: [Interval constant: 0.0]];
    [eyield2 set: [Interval constant: 0.0]];
    if([FearlusArguments debugInnerCoreMethods]) {
      printf("i: Calculation of Expected Yield:\n\tYield = %u, antimatch = %u,"
	     " land use bits = %u\n", yield, antimatch, lubits);
      printf("\tYield Yield^2        No. of ways             Out of   "
	     "Probability\n");
    }
    for(i = 0; i <= (int)yield; i++) {
      Interval *xi = [[Interval create]
		       set:
			 [Interval
			   castInt: ((int)yield + (int)antimatch - (2 * i))]];
      Interval *pi = [[Interval create]
		       set:
			 [
			  [Interval
			    castULL: [MiscFunc combineN: (int)antimatch r: i]
			  ]
			   prod:
			     [Interval
			       castULL: [MiscFunc
					  combineN: (int)(lubits - antimatch)
					  r: (int)yield - i]
			     ]
			 ]];

      [eyield add: [xi prod: pi]];
      [eyield2 add: [[xi prod: xi] prod: pi]];
      if([FearlusArguments debugInnerCoreMethods]) {
	printf("\t%s %s %s %s %s\n",
	       [xi printfVar: "% 5g"],
	       [[xi prod: xi] printfVar: "% 7g"],
	       [pi printfVar: "% 18g"],
	       [total printfVar: "% 18g"],
	       ([total gt: [Interval constant: 0.0]]
		? [[xi quot: total] printfVar: "% 13g"]
		: [[Interval constant: 0.0] printfVar: "% 13g"]));
	[Interval printfclean];
      }
      [xi drop];
      [pi drop];
    }
    [eyield div: total];
    [eyield2 div: total];

    [var set: [eyield2 diff: [eyield prod: eyield]]];

    [eyield2 drop];
    [total drop];
    return [eyield ret];
  }

  tempZone = [Zone create: scratchZone];
  bucket = [SelectUseBucket create: tempZone
			    withParameters: parameter
			    andLandUses: [[lp getEnvironment] getLandUses]];
  vararr = [AssocArray createBegin: tempZone];
  [vararr setSize: [parameter nUses]];
  vararr = [vararr createEnd];
  // Loop through land parcels owned by the land manager
  if([FearlusArguments debugCoreMethods]) {
    printf("S: %s: Using imitative optimum strategy of land manager %u to "
	   "determine land use for LP = land parcel %u at (%d, %d)\n",
	   sel_get_name(_cmd), [lm getPIN], [lp getPIN], [lp getX], [lp getY]);
  }
  
  for(parcelIndex = [[lm getLandParcels] begin: tempZone], [parcelIndex next];
      [parcelIndex getLoc] == Member;
      [parcelIndex next]) {
    Interval *score = [Interval create];
    Interval *variance = [Interval create];
    Number *n;

    myParcel = [parcelIndex get];
    yield = [myParcel getLastYield];
    lu = [myParcel getLandUse];
    match = [[lp getAttributes] getMatchWith: [myParcel getAttributes]];
    [score set: getExpectedYieldVar(yield,
				    [parameter parcelBitStringSize] - match,
				    [parameter useBitStringSize], variance)];
    n = [vararr getObjectWithKey: lu];
    if(n == nil) {
      [bucket setScore: [score call] inBucketForLandUse: lu];
      [myParcel incNImitations];
      n = [Number create: tempZone];
      [n setIDouble: [variance call]];
      if(![vararr addObject: n withKey: lu]) {
	fprintf(stderr, "CautiousImitativeOptimumStrategy: Adding variance %s "
		"of land use %u to the associative array failed\n",
		[variance printf: "%g"], [lu getPIN]);
	[Interval printfclean];
	abort();
      }
      if([FearlusArguments debugCoreMethods]) {
	printf("S: %s: Found new land use %u at land parcel %u at (%d, %d) "
	       " with antimatch to LP of %u and yield %u. Expected yield "
	       " (= score for LU) = %s. Variance = %s\n", sel_get_name(_cmd),
	       [lu getPIN], [myParcel getPIN], [myParcel getX],
	       [myParcel getY], [parameter parcelBitStringSize] - match, 
	       yield, [score printf: "%g"], [variance printf: "%g"]);
	[Interval printfclean];
      }
    }
    else if([variance lt: [n getIDouble]]) {
      [bucket setScore: [score call] inBucketForLandUse: lu];
      [n setIDouble: [variance call]];
      if([FearlusArguments debugCoreMethods]) {
	printf("S: %s: Found expected yield (%s) for land use %u with lower "
	       "variance (%s) at land parcel %u at (%d, %d).\n",
	       sel_get_name(_cmd), [score printf: "%g"], [lu getPIN],
	       [variance printf: "%g"],
	       [myParcel getPIN], [myParcel getX], [myParcel getY]);
	[Interval printfclean];
      }
    }
    else {
      if([FearlusArguments debugCoreMethods]) {
	printf("S: %s: Ignoring expected yield (%s) for land use %u with "
	       "higher variance (%s) at land parcel %u at (%d, %d).\n",
	       sel_get_name(_cmd), [score printf: "%g"], [lu getPIN],
	       [variance printf: "%g"],
	       [myParcel getPIN], [myParcel getX], [myParcel getY]);
	[Interval printfclean];
      }
    }
    [score drop];
    [variance drop];

  }
  [parcelIndex drop];
  // Loop through the neighbours, and through their land parcels
  bs = [BitString create: tempZone
		  setBitCount: [parameter parcelBitStringSize]];
  for(i = 0; (neighbour = [lm getNextNeighbour: i]); i |= 1) {
    if([FearlusArguments debugCoreMethods]) {
      printf("S: %s: Adding scores for land uses of land parcels belonging"
	     " to land manager %u\n", sel_get_name(_cmd), [neighbour getPIN]);
    }
    for(parcelIndex = [[neighbour getLandParcels] begin: tempZone],
	  [parcelIndex next];
	[parcelIndex getLoc] == Member;
	[parcelIndex next]) {
      Interval *score = [Interval create];
      Interval *variance = [Interval create];
      Number *n;

      neighbouringParcel = [parcelIndex get];
      lu = [neighbouringParcel getLandUse];
      [bs setToCopyOf: [neighbouringParcel getAttributes]];
      [bs mutateWithProbability: [lm getNeighbourNoise]];
      match = [[lp getAttributes] getMatchWith: bs];
      yield = [neighbouringParcel getLastYield];
      [score set: getExpectedYieldVar(yield,
				      [parameter parcelBitStringSize] - match,
				      [parameter useBitStringSize], variance)];
      n = [vararr getObjectWithKey: lu];
      if(n == nil) {
	[bucket setScore: [score call] inBucketForLandUse: lu];
	[neighbouringParcel incNImitations];
	n = [Number create: tempZone];
	[n setIDouble: [variance call]];
	if(![vararr addObject: n withKey: lu]) {
	  fprintf(stderr, "CautiousImitativeOptimumStrategy: Adding variance"
		  " %s of land use %u to the associative array failed\n",
		  [variance printf: "%g"], [lu getPIN]);
	  [Interval printfclean];
	  abort();
	}

	if([FearlusArguments debugCoreMethods]) {
	  printf("S: %s: Found new land use %u at neighbour's land parcel %u "
		 "at (%d, %d) with antimatch to LP of %u and yield %u"
		 "Expected yield = %s. Variance = %s.\n",
		 sel_get_name(_cmd),
		 [lu getPIN], [neighbouringParcel getPIN],
		 [neighbouringParcel getX], [neighbouringParcel getY],
		 [parameter parcelBitStringSize] - match, yield,
		 [score printf: "%g"], [variance printf: "%g"]);
	  [Interval printfclean];
	}
      }
      else if([variance lt: [n getIDouble]]) {
	[bucket setScore: [score call] inBucketForLandUse: lu];
	[n setIDouble: [variance call]];
	if([FearlusArguments debugCoreMethods]) {
	  printf("S: %s: Found expected yield (%s) for land use %u with lower "
		 "variance (%s) at neighbouring land parcel %u at (%d, %d).\n",
		 sel_get_name(_cmd), [score printf: "%g"], [lu getPIN],
		 [variance printf: "%g"],
		 [neighbouringParcel getPIN],
		 [neighbouringParcel getX], [neighbouringParcel getY]);
	  [Interval printfclean];
	}
      }
      else {
	if([FearlusArguments debugCoreMethods]) {
	  printf("S: %s: Ignoring expected yield (%s) for land use %u with "
		 "higher variance (%s) at neighbouring parcel %u at (%d, %d)."
		 "\n", sel_get_name(_cmd), [score printf: "%g"], [lu getPIN],
		 [variance printf: "%g"],
		 [neighbouringParcel getPIN],
		 [neighbouringParcel getX], [neighbouringParcel getY]);
	  [Interval printfclean];
	}
      }
      [score drop];
      [variance drop];
    }
    [parcelIndex drop];
  }
  lu = [bucket getRandomOptimumChoice];
  [bucket drop];
  [bs drop];
  [tempZone drop];
  return lu;
}

@end
