/*
    FEARLUS model0-7-1: Same10PropClumper.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1, an agent-based model of land use
    change.

    FEARLUS model0-7-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


The algorithm is as follows:

For each bit:
1  Make a list ZERO of all the cells with value 0, in random order

2  Make a list ONES of all the cells with value 1, in random order

3  Let A = the first item of ZERO

4  Search in ONES for a cell B to swap with that
     matches a criterion saying that:
       One layout L1 is more clumped than another L2 if:
         (a) the number of von Neumann boundaries between cells with both cells
	     having the same value in L1 is greater than that in L2
	 OR if (a) for L1 = that for L2, then if
	 (b) the number of diagonal boundaries between cells with both cells
	     having the same value in L1 is greater than that in L2

5  If the search is successful, then set A's bit to 1, B's bit to 0, put A
   onto the end of ONES, and B onto the end of ZERO, and start again from 3

6  If the search is unsuccessful, then let A be the next item of ZERO

7  Repeat from 4 if A exists (i.e. we weren't at the end of ZERO in the last
   iteration)

*/

#import "Same10PropClumper.h"
#import "Environment.h"
#import "LandParcel.h"
#import "FearlusArguments.h"
#import "MiscFunc.h"
#import "Parameter.h"
#import <collections/List.h>
#import <stdio.h>

/*

computeBitTotal(LandParcel *, int, int *, int *, int *, int *)

Compute the number of diagonal and von Neumann neighbours of the land parcel
containing bit value 1 or 0. The behaviour is toroidal.

The function is static to indicate that it can only be used in this file.

*/

static void computeBitTotal(LandParcel *lp,
			    int bit,
			    int *nZeroVonNeumann,
			    int *nZeroDiagonal,
			    int *nOneVonNeumann,
			    int *nOneDiagonal) {
  int x, y, xx, yy, maxx, maxy, lpx, lpy;
  Environment *env;

  env = [lp getEnvironment];

  maxx = [env getSizeX];
  maxy = [env getSizeY];

  lpx = [lp getX];
  lpy = [lp getY];

  (*nZeroVonNeumann) = (*nZeroDiagonal) = (*nOneVonNeumann) = (*nOneDiagonal)
    = 0;
  for(x = -1; x <= 1; x++) {
    for(y = -1; y <= 1; y++) {
      int bitValue;

      if(x == 0 && y == 0) continue;	// Don't include yourself in the count!

      xx = [MiscFunc Mod: (lpx + x) inRange0To: maxx];
      yy = [MiscFunc Mod: (lpy + y) inRange0To: maxy];

      if(xx == lpx && yy == lpy) continue;
                                        /* Don't include yourself in the count!
					   (Even when wrapping-around) */

      bitValue = [[env getObjectAtX: xx Y: yy] getOneAttribute: bit];

      if(x == 0 || y == 0) {		// Von Neumann neighbours
	if(bitValue == 1) {
	  (*nOneVonNeumann)++;
	}
	else {
	  (*nZeroVonNeumann)++;
	}
      }
      else if(x == y || x == -y) {	// Diagonal neighbours
	if(bitValue == 1) {
	  (*nOneDiagonal)++;
	}
	else {
	  (*nZeroDiagonal)++;
	}
      }
    }
  }
}

/*

nVonNeumannBoundaries(lp1, lp2)

Return the number of von Neumann boundaries shared by lp1 and lp2 (including
wrapped-around boundaries)

*/

static int nVonNeumannBoundaries(LandParcel *lp1, LandParcel *lp2) {
  int x1, y1, x2, y2, xE, yE, x, y, nB = 0;

  x1 = [lp1 getX];
  y1 = [lp1 getY];
  x2 = [lp2 getX];
  y2 = [lp2 getY];

  xE = [[lp1 getEnvironment] getSizeX];
  yE = [[lp1 getEnvironment] getSizeY];

  // Search Moore Neighbourhood of x1 for neighbours 
  for(x = -1; x <= 1; x++) {
    for(y = -1; y <= 1; y++) {
      if(x == 0 && y == 0) continue;	// ignore the centre, obviously
      if(x * y != 0) continue;		// ignore Diagonals
      if([MiscFunc Mod: (x1 + x) inRange0To: xE] == x2
	 && [MiscFunc Mod: (y1 + y) inRange0To: yE] == y2) {
	nB++;
      }
    }
  }

  return nB;
}

/*

nDiagonalBoundaries(lp1, lp2)

Return the number of diagonal boundaries shared by lp1 and lp2 (including
wrapped-around boundaries)

*/

static int nDiagonalBoundaries(LandParcel *lp1, LandParcel *lp2) {
  int x1, y1, x2, y2, xE, yE, x, y, nB = 0;

  x1 = [lp1 getX];
  y1 = [lp1 getY];
  x2 = [lp2 getX];
  y2 = [lp2 getY];

  xE = [[lp1 getEnvironment] getSizeX];
  yE = [[lp1 getEnvironment] getSizeY];

  // Search Moore Neighbourhood of x1 for neighbours 
  for(x = -1; x <= 1; x++) {
    for(y = -1; y <= 1; y++) {
      if(x == 0 && y == 0) continue;	// ignore the centre, obviously
      if(x * y == 0) continue;		// ignore von Neumann neighbours
      if([MiscFunc Mod: (x1 + x) inRange0To: xE] == x2
	 && [MiscFunc Mod: (y1 + y) inRange0To: yE] == y2) {
	nB++;
      }
    }
  }

  return nB;
}

@implementation Same10PropClumper

/*

create:

Create the object and initialise its instance variables

*/

+create: aZone {
  Same10PropClumper *obj;

  obj = [super create: aZone];
  obj->maxBitSpecified = NO;

  return obj;
}

/*

setParameters:

Pass the parameter object into the clumper

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
}

/*

setParameter:toValue:

This is a means of passing in various clump-class specific parameters into the
clump object. These will be called from the Parameter object, I guess.

*/

-(BOOL)setParameter: (char *)param toValue: (char *)value {
  if(strcmp(param, "bitsToClump") == 0) {
    maxBit = (unsigned)atoi(value);
    maxBitSpecified = YES;
    return YES;
  }
  else {
    fprintf(stderr, "Invalid parameter for Same10PropClumper (%s)\n"
	    "Same10PropClumper has the following parameters:\n"
	    "\tbitsToClump\t\tSpecify the first n bits to clump\n", param);
    return NO;
  }
}

/*

clumpEnvironment:

Clump the environment using the algorithm described at the top of the file. A
practical implementation of the clumping criterion would be something like:

Let cvnA = the number of A's von Neumann neighbours set to 0
Let cdA = the number of A's diagonal neighbours set to 0

Let cvnB = the number of B's von Neumann neighbours set to 1
Let cdB = the number of B's diagonal neighbours set to 1

Let ovnA = the number of A's von Neumann neighbours set to 1
Let odA = the number of A's diagonal neighbours set to 1

Let ovnB = the number of B's von Neumann neighbours set to 0
Let odB = the number of B's diagonal neighbours set to 0

We need to choose B such that 

(ovnA + ovnB) - (cvnA + cvnB) is maximised and more than zero. Where there
are more than 1 such choices for B, then maximise

(odA + odB) - (cdA + cdB)

*/

-(void)clumpEnvironment: (Environment *)env {
  id <List> zero, ones;
  id <List> tmpOnesList, tmpZeroList;
  unsigned bit;
  LandParcel *lp;
  id <Zone> clumpyZone;

  if([FearlusArguments debugCoreMethods]) {
    printf("E: %s: Clumping (Same10PropClumper)"
	   "\n", sel_get_name(_cmd));
  }
  else {
    printf("Clumping (Same10PropClumper) "
	   "(1 dot for each item on the zero list) ");
  }
  fflush(stdout);
  clumpyZone = [Zone create: scratchZone];

  zero = [List create: clumpyZone];
  ones = [List create: clumpyZone];
  tmpOnesList = [List create: clumpyZone];
  tmpZeroList = [List create: clumpyZone];
  if(!maxBitSpecified) maxBit = [parameter parcelBitStringSize];
  else if(maxBit > [parameter parcelBitStringSize]) {
    fprintf(stderr, "WARNING: bitsToClump parameter of Same10PropClumper is "
	    "set to more than the number of bits in the land parcel bit "
	    "string. (%u, %u)\n", maxBit, [parameter parcelBitStringSize]);
    maxBit = [parameter parcelBitStringSize];
  }
  for(bit = 0; bit < maxBit; bit++) {
    int x, y, swapCount;

    if([FearlusArguments debugCoreMethods]) {
      [env printTextParcelBitstringsBit: bit];
    }
    /* Create the lists of those with 1 and those with 0 in random order */
    for(x = 0; x < [env getSizeX]; x++) {
      for(y = 0; y < [env getSizeY]; y++) {
	id <List> zeroOrOne;
	lp = [env getObjectAtX: x Y: y];
	if([lp getOneAttribute: bit] == 0) {
	  zeroOrOne = zero;
	}
	else {
	  zeroOrOne = ones;
	}
	/* Make sure the list is in random order */
	if([uniformDblRand getDoubleWithMin: 0.0 withMax: 1.0] < 0.5) {
	  [zeroOrOne addLast: lp];
	}
	else {
	  [zeroOrOne addFirst: lp];
	}
      }
    }

    do {
      swapCount = 0;
      while([zero getCount] > 0) {
	int zvs;		/* Number of [v]on Neumann neighbours with the
				   [s]ame value as a member of the [z]ero 
				   array */
	int zvd;		/* Number of [v]on Neumann neighbours with a
				   [d]ifferent value from a member of the
				   [z]ero array	*/
	int zds;		/* [d]iagonal neighbours, [s]ame value,
				   [z]ero */
	int zdd;		/* [d]iagonal, [d]ifferent, [z]ero */
	int ovs, ovd, ods, odd;	/* Ditto for [o]nes array */
	LandParcel *maxClumpyLP;
				/* The maximum clumpiness land
				   parcel. This will be the land
				   parcel which, when the bits are
				   swapped with the current land
				   parcel, will result in the maximum
				   clumpiness. */
	int maxClumpyVonN;	/* The von Neumann clumpiness of the
				   maximum clumpiness land parcel. */
	int maxClumpyDiag;	/* The diagonal clumpiness of the
				   maximum clumpiness land parcel. */
	LandParcel *onesLP;	/* A land parcel from the ones list */
      
	lp = [zero removeFirst];
	computeBitTotal(lp, bit, &zvs, &zds, &zvd, &zdd);
	maxClumpyVonN = maxClumpyDiag = 0;
	maxClumpyLP = nil;

	/* Search through the ones list to find the land parcel to swap with
	   that makes the maximum improvement in clumpiness. At the same time,
	   we want to keep out the land parcel to swap with, so we're not
	   messing about trying to look for it. So use tmpOnesList to store all
	   the land parcels we've rejected, and then we'll put them back after.
	*/
	while([ones getCount] > 0) {
	  int currClumpy;
	  int swapClumpy;
	  LandParcel *tmpLP;
	  
	  onesLP = [ones removeFirst];
	  computeBitTotal(onesLP, bit, &ovd, &odd, &ovs, &ods);

	  currClumpy = zvs + ovs;  	// Clumpiness if we leave as is
	  swapClumpy = zvd + ovd 
	    - (2 * nVonNeumannBoundaries(lp, onesLP));
                                   	// Clumpiness if we swap zero & one

	  if([FearlusArguments debugInnerCoreMethods]) {
	    printf("Considering parcel %u at (%d, %d) and parcel %u at (%d, "
		   "%d): Von Neumann clumpy score if leave: %d, if swap: %d,"
                   " max: %d.", [lp getPIN], [lp getX], [lp getY],
		   [onesLP getPIN], [onesLP getX], [onesLP getY], currClumpy,
		   swapClumpy, maxClumpyVonN);
	    printf("\nzvs = %d, zds = %d, zvd = %d, zdd = %d\n"
		   "ovd = %d, odd = %d, ovs = %d, ods = %d\n"
		   "VN nbrs = %d, D nbrs = %d",
		   zvs, zds, zvd, zdd, ovd, odd, ovs, ods,
		   nVonNeumannBoundaries(lp, onesLP),
		   nDiagonalBoundaries(lp, onesLP));
	  }
	  tmpLP = onesLP;
	  if(swapClumpy > currClumpy) {	// Clumpier if we swap
	    if(swapClumpy > maxClumpyVonN) {
	                                // Clumpier than the best so far
	      maxClumpyVonN = swapClumpy;
	      tmpLP = maxClumpyLP;
	      maxClumpyLP = onesLP;
	      maxClumpyDiag = zdd + odd;
	      if([FearlusArguments debugInnerCoreMethods]) {
		printf(" Best.");
	      }
	    }
	    else if(swapClumpy == maxClumpyVonN) {
	                                // Just as von Neumann clumpy ...
	      swapClumpy = zdd + odd 
		- (2 * nDiagonalBoundaries(lp, onesLP));
	      currClumpy = zds + ods;
	      if([FearlusArguments debugInnerCoreMethods]) {
		printf(" Diagonal clumpy score if leave: %d, if swap: %d, "
		       "max: %d.", currClumpy, swapClumpy, maxClumpyDiag);
	      }
	      if((swapClumpy > currClumpy) && (swapClumpy > maxClumpyDiag)) {
	                                // ... but more diagonal clumpy
		tmpLP = maxClumpyLP;
		maxClumpyLP = onesLP;
		maxClumpyDiag = swapClumpy;
		if([FearlusArguments debugInnerCoreMethods]) {
		  printf(" Best.");
		}
	      }
	    }
	  }
	  if([FearlusArguments debugInnerCoreMethods]) {
	    printf("\n");
	  }
	  if(tmpLP != nil) [tmpOnesList addLast: tmpLP];
	}
	/* maxClumpyLP now contains the land parcel which, if the bits
	   are swapped, will result in the maximum increase in
	   clumpiness, if maxClumpyVonN > 0 or maxClumpyDiag > 0 */
	if(maxClumpyVonN > 0 || maxClumpyDiag > 0) {
	  /* Swap the one and zero to make things more clumpy */
	  [lp setAttribute: bit toValue: 1];
	  [maxClumpyLP setAttribute: bit toValue: 0];
	  [tmpOnesList addLast: lp];
	  [tmpZeroList addLast: maxClumpyLP];
	  swapCount++;
	  if([FearlusArguments debugInnerCoreMethods]) {
	    printf("Swapping parcel %u at (%d, %d) and %u at (%d, %d)\n",
		   [lp getPIN], [lp getX], [lp getY], [maxClumpyLP getPIN],
		   [maxClumpyLP getX], [maxClumpyLP getY]);
	  }
	}
	else {
	  [tmpZeroList addLast: lp];
	}
	/* Restore the ones list to its former glory */
	while([tmpOnesList getCount] > 0) {
	  [ones addLast: [tmpOnesList removeFirst]];
	}
      }
      /* Print progress information */
      if([FearlusArguments debugCoreMethods]) {
	[env printTextParcelBitstringsBit: bit];
      }
      else {
	printf(".");
	fflush(stdout);
      }
      /* Restore the zero list to its former glory */
      while([tmpZeroList getCount] > 0) {
	[zero addFirst: [tmpZeroList removeLast]];
      }
    } while(swapCount > 0);
    /* Clean up ready for the next bit */
    [ones removeAll];
    [zero removeAll];
  }
  [clumpyZone drop];
  printf("\n");
  printf("Done Clumping\n");
}

@end
