/*
    FEARLUS model0-7-1: SubPopulation.m
    Copyright (C) 1999-2003, 2006  Macaulay Institute

    This file is part of FEARLUS model0-7-1, an agent-based model of land use
    change.

    FEARLUS model0-7-1 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-7-1 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the SubPopulation object. From model 0-3 onwards,
this object has been provided so that you can have specific
sub-populations with separate sets of values for each of the
strategy-related parameters. The collective term for a set of strategy
types defined over the four contexts (above imitation threshold, below
imitation threshold imitative and non-imitative and initial)
parameterised by an aspiration threshold, an imitation probability and
a neighbourhood weighting is a "DECISION ALGORITHM".

Enhancements for model 0-5-3:

Include capability of specifying a normal rather than uniform distribution
for each of the subpopulation parameter ranges.

*/

#import "SubPopulation.h"
#import "StrategySelector.h"
#import "LandManager.h"
#import "Parameter.h"
#import "FearlusArguments.h"
#import "ClassInfo.h"
#import "Interval.h"
#import <random.h>
#import <objc/objc-api.h>

@implementation SubPopulation

/*

create:

Create the sub-population, assigning some default values and creating
the PIN.  Create a probemap for the sub-population class, which
includes the clickMeToGetStrategySelectorGUI method.

*/

+create: aZone {
  SubPopulation *sp;
  static unsigned counter = 1;
  id <ProbeMap> pm;

  sp = [super create: aZone];
  sp->pin = counter;
  sp->nRemovals = 0;
  sp->probability = 0.5;
  sp->landMgrClass = "LandManager";
  sp->neighbourWeightMax = 1.0;
  sp->neighbourWeightMin = 1.0;
  sp->neighbourWeightDist = "uniform";
  sp->neighbourWeightMean = 1.0;
  sp->neighbourWeightVar = 0.0;
  sp->imitateProbMax = 0.75;
  sp->imitateProbMin = 0.5;
  sp->imitateProbDist = "uniform";
  sp->imitateProbMean = 0.625;
  sp->imitateProbVar = 0.125;
  sp->habitImitateThresholdMax = 14.0; // aspiration
  sp->habitImitateThresholdMin = 10.0; // aspiration
  sp->habitImitateThresholdDist = "uniform"; // aspiration
  sp->habitImitateThresholdMean = 12.0;	// aspiration
  sp->habitImitateThresholdVar = 2.0; // aspiration
  sp->memorySizeMin = 1;
  sp->memorySizeMax = 1;
  sp->colour = (int)counter;
  sp->strategySelectorFile = "fearlus.st";
  sp->label = (char *)[aZone alloc: (size_t)12];
  // 12 = "SubPop-XXX\0" + 1 for luck (XXX allows for 3 digit PIN)
  sprintf(sp->label, "SubPop % 3d", counter % 1000);
  // We can have a more than 3 digit PIN, but it won't appear on the label. I'd
  // like to see a simulation with 1000 sub-populations anyway! Imagine the
  // button panel!
  sp->selectorCreated = YES;
  sp->strategies = [StrategySelector create: aZone];
  sp->managerList = [List create: aZone];
  sp->normalDist = [NormalDist create: aZone setGenerator: randomGenerator];

  pm = [EmptyProbeMap createBegin: aZone];
  [pm setProbedClass: [self class]];
  pm = [pm createEnd];

  [pm addProbe: [probeLibrary getProbeForVariable: "probability"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "landMgrClass"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightMin"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightMax"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightDist"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightMean"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightVar"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbMin"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbMax"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbDist"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbMean"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbVar"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdMin"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdMax"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdDist"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdMean"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdVar"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "memorySizeMin"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "memorySizeMax"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "strategySelectorFile"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForMessage: 
				"clickMeToGetStrategySelectorGUI"
			      inClass: [self class]]];
  [probeLibrary setProbeMap: pm For: [self class]];

  counter++;
  return(sp);
}

/*

getPIN

Return the unique PIN assigned to the sub-population at create time.

*/

-(unsigned)getPIN {
  return pin;
}

/*

getLabel

Return the label (created from the PIN at create time).

*/

-(char *)getLabel {
  return label;
}

/*

setColour:

Set the colour to be used to represent this sub-population

*/

-(void)setColour: (int)col {
  colour = col;
}

/*

getColour

Get the colour for the sub-population

*/

-(int)getColour {
  return colour;
}

/*

getNRemovals

Return the number of land managers removed from this subpopulation since
resetRemovals was called.

*/

-(unsigned)getNRemovals {
  return nRemovals;
}

/*

resetRemovals

Reset the number of removals to zero.

*/

-resetRemovals {
  nRemovals = 0;
  return self;
}

/*

getLandManagerClass

Return the class of land manager to use for this subpopulation.

*/

-(Class)getLandManagerClass {
  return objc_get_class(landMgrClass);
}

/*

getANeighbourWeight

Return a randomly generated neighbourhood weighting parameter. This is a method
that will be called from land managers when they are created.

*/

-(Interval *)getANeighbourWeight {
  // Don't bother generating random numbers if min is equal to max
  if(strcmp(neighbourWeightDist, "uniform") == 0) {
    if(neighbourWeightMin == neighbourWeightMax)
      return [[Interval constant: neighbourWeightMin] ret];
    return [[Interval constant:
		       [uniformDblRand getDoubleWithMin: neighbourWeightMin
				       withMax: neighbourWeightMax]] ret];
  }
  else {
    if(neighbourWeightVar == 0.0)
      return [[Interval constant: neighbourWeightMean] ret];
				// ... or if variance is zero
    return [[Interval constant:
		       [normalDist getSampleWithMean: neighbourWeightMean
				   withVariance: neighbourWeightVar]] ret];
  }
}

/*

getNeighbourWeightMin

Return the minimum neighbourhood weighting

*/

-(double)getNeighbourWeightMin {
  return neighbourWeightMin;
}

/*

getNeighbourWeightMax

Return the maximum neighbourhood weighting

*/

-(double)getNeighbourWeightMax {
  return neighbourWeightMax;
}

/*

getNeighbourWeightMean

Return the mean neighbourhood weighting

*/

-(double)getNeighbourWeightMean {
  return neighbourWeightMean;
}

/*

getNeighbourWeightVar

Return the variance of the neighbourhood weighting

*/

-(double)getNeighbourWeightVar {
  return neighbourWeightVar;
}

/*

getNeighbourWeightDist

Return the distribution used to generate neighbourhood weightings

*/

-(char *)getNeighbourWeightDist {
  return neighbourWeightDist;
}

/*

getAnImitateProb

Return a randomly generated imitative strategy probability, within the range
specified for this sub-population. This is another method land managers will
call at create time.

*/

-(Interval *)getAnImitateProb {
  if(strcmp(imitateProbDist, "uniform") == 0) {
    if(imitateProbMin == imitateProbMax)
      return [[Interval constant: imitateProbMin] ret];
    return [[Interval constant:
			[uniformDblRand getDoubleWithMin: imitateProbMin
					withMax: imitateProbMax]] ret];
  }
  else {
    if(imitateProbVar == 0.0)
      return [[Interval constant: imitateProbVar] ret];
    return [[Interval constant:
			[normalDist getSampleWithMean: imitateProbMean
				    withVariance: imitateProbVar]] ret];
  }
}

/*

getImitateProbMin

Return the minimum imitate probability

*/

-(double)getImitateProbMin {
  return imitateProbMin;
}


/*

getImitateProbMax

Return the maximum imitate probability

*/

-(double)getImitateProbMax {
  return imitateProbMax;
}

/*

getImitateProbMean

Return the mean imitation probability

*/

-(double)getImitateProbMean {
  return imitateProbMean;
}

/*

getImitateProbVar

Return the variance of the imitation probability

*/

-(double)getImitateProbVar {
  return imitateProbVar;
}

/*

getImitateProbDist

Return the distribution used to generate imitation probabilities

*/

-(char *)getImitateProbDist {
  return imitateProbDist;
}

/*

getAHabitImitateThreshold (aspiration)

Return a randomly generated aspiration threshold, within the range
specified for this sub-population. This is another method land
managers will call at create time.

*/

-(Interval *)getAHabitImitateThreshold { // aspiration
  /* throughout this method, "habitImitate" refers to "aspiration" */
  if(strcmp(habitImitateThresholdDist, "uniform") == 0) {
    if(habitImitateThresholdMax == habitImitateThresholdMin) {
      return [[Interval constant: habitImitateThresholdMin] ret];
    }
    return [[Interval constant:
			[uniformDblRand getDoubleWithMin:
					  habitImitateThresholdMin
					withMax: habitImitateThresholdMax]]
	     ret];
  }
  else {
    if(habitImitateThresholdVar == 0.0)
      return [[Interval constant: habitImitateThresholdMean] ret];
    return [[Interval constant:
			[normalDist getSampleWithMean:
				      habitImitateThresholdMean
				    withVariance: 
				      habitImitateThresholdVar]] ret];
  }
}

/*

getHabitThresholdMin

Return the minimum aspiration threshold

*/

-(double)getHabitThresholdMin {
  return habitImitateThresholdMin; // aspiration
}

/*

getHabitThresholdMax

Return the maximum aspiration threshold

*/

-(double)getHabitThresholdMax {
  return habitImitateThresholdMax; // aspiration
}

/*

getHabitThresholdMean

Return the mean aspiration threshold

*/

-(double)getHabitThresholdMean {
  return habitImitateThresholdMean; // aspiration
}

/*

getHabitThresholdVar

Return the variance of the aspiration threshold

*/

-(double)getHabitThresholdVar {
  return habitImitateThresholdVar; // aspiration
}

/*

getHabitThresholdDist

Return the distribution used to generate aspiration thresholds

*/

-(char *)getHabitThresholdDist {
  return habitImitateThresholdDist; // aspiration
}

/*

getAMemorySize

Return a randomly generated memory size within the range specified for this
sub-population.

*/

-(unsigned)getAMemorySize {
  if(memorySizeMax == memorySizeMin) {
    return memorySizeMin;
  }
  return [uniformUnsRand getUnsignedWithMin: memorySizeMin
			 withMax: memorySizeMax];
}

/*

getMaxMemorySize

Return the maximum memory size of this subpopulation

*/

-(unsigned)getMaxMemorySize {
  return memorySizeMax;
}

/*

addLandManager:

Add the land manager to the list of land managers belonging to this sub
population

*/

-(void)addLandManager: (LandManager *)lm {
  [managerList addLast: lm];
}

/*

removeLandManager:

Remove the land manager from the list of land managers belonging to this
sub-population

*/

-(void)removeLandManager: (LandManager *)lm {
  nRemovals++;
  [managerList remove: lm];
}

/*

getLandManagers

Return the list of land managers belonging to this subpopulation

*/

-(id <List>)getLandManagers {
  return managerList;
}

/*

getAnAboveStrategyForLandManager:andParameters:

Return a randomly generated contentment strategy from the strategy
selector object.

*/

-(id <Strategy>)getAnAboveStrategyForLandManager: (LandManager *)lm
				   andParameters: (Parameter *)p {
  id <Strategy> strategy;

  strategy = [[strategies selectAboveStrategy]
	       create: [self getZone]
	       withManager: lm
	       andParameters: p];
  return strategy;
}

/*

getABelowNonImitativeStrategyForLandManager:andParameters:

Return a randomly selected non-imitative strategy from the strategy selector
object for the specified land manager and parameters

*/

-(id <NonImitativeStrategy>)
     getABelowNonImitativeStrategyForLandManager: (LandManager *)lm 
				   andParameters: (Parameter *)p {
  id <NonImitativeStrategy> strategy;

  strategy = (id <NonImitativeStrategy>)
    [[strategies selectBelowNonImitativeStrategy]
      create: [self getZone]
      withManager: lm
      andParameters: p];
  return strategy;
}

/*

getABelowImitativeStrategyForLandManager:andParameters:

Return a randomly selected imitative strategy from the strategy selector object
for the specified land manager and parameters.

*/

-(id <ImitativeStrategy>)
     getABelowImitativeStrategyForLandManager: (LandManager *)lm 
				andParameters: (Parameter *)p {
  id <ImitativeStrategy> strategy;

  strategy = (id <ImitativeStrategy>)
    [[strategies selectBelowImitativeStrategy]
      create: [self getZone]
      withManager: lm
      andParameters: p];
  return strategy;
}

/*

getAnInitialStrategyForLandManager:andParameters:

Return a randomly selected initial strategy from the strategy selector object
for the specified land manager and parameters.

*/

-(id <NonImitativeStrategy, NonHistoricalStrategy>)
     getAnInitialStrategyForLandManager: (LandManager *)lm 
			  andParameters: (Parameter *)p {
  id <NonImitativeStrategy, NonHistoricalStrategy> strategy;

  strategy = (id <NonImitativeStrategy, NonHistoricalStrategy>)
    [[strategies selectInitialStrategy]
      create: [self getZone]
      withManager: lm
      andParameters: p];
  return strategy;
}

/*

getProb

Return the probability that a land manager will be created that belongs to this
sub-population

*/

-(Interval *)getProb {
  return [[Interval constant: probability] ret];
}

/*

loadFromFileNamed:

Load certain parameters of this object from the specified
file. Specifically, the prob, landMgrClass, colour, neighbourWeightMax,
neighbourWeightMin, imitateProbMax, imitateProbMin,
habitImitateThresholdMax, habitImitateThresholdMin and
strategySelectorFile may all be loaded in, as can the lable. The PIN
may not.

This method gets called from the Parameter object if the user has not
done any GUI strategy parameter setting. So, the strategy selector
object should also get loaded in here.

Nevertheless, we'll check whether the strategy selector object has
been created before possibly recreating it through loading it from a
file.

*/
			    
-loadFromFileNamed: (const char *)filename {
  BOOL dummySelectorCreated = selectorCreated;
  unsigned dummyPin = pin;
  Class lmClass;
  // Keep copies of the variables that may not be loaded.

  [ObjectLoader load: self fromFileNamed: filename];

  lmClass = [self getLandManagerClass];
  if(![ClassInfo class: lmClass isSubClassOf: [LandManager class]]) {
    fprintf(stderr, "Land Manager class %s is not (a subclass of) LandManager "
	    "in sub-pop file %s\n", class_get_class_name(lmClass), filename);
    abort();
  }
  selectorCreated = dummySelectorCreated;
  pin = dummyPin;
  
  if(selectorCreated) {
    [strategies drop];
  }
  strategies = [StrategySelector create: [self getZone]];
  [strategies loadFromFileNamed: strategySelectorFile];

  selectorCreated = YES;
  return self;
}

/*

saveToFileNamed:

Use the ObjectSaver class to save the sub-population data to the specified
filename. Before that, however, create a probemap showing which instance
variables are to be saved.

*/

-saveToFileNamed: (const char *)filename {
  id <ProbeMap> pm;

  pm = [EmptyProbeMap createBegin: [self getZone]];
  [pm setProbedClass: [self class]];
  pm = [pm createEnd];

  [pm addProbe: [probeLibrary getProbeForVariable: "landMgrClass"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightDist"
			      inClass: [self class]]];
  if(strcmp(neighbourWeightDist, "uniform") == 0) {
    [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightMin"
				inClass: [self class]]];
    [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightMax"
				inClass: [self class]]];
  }
  else {
    [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightMean"
				inClass: [self class]]];
    [pm addProbe: [probeLibrary getProbeForVariable: "neighbourWeightVar"
				inClass: [self class]]];
  }

  [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbDist"
			      inClass: [self class]]];
  if(strcmp(imitateProbDist, "uniform") == 0) {
    [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbMin"
				inClass: [self class]]];
    [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbMax"
				inClass: [self class]]];
  }
  else {
    [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbMean"
				inClass: [self class]]];
    [pm addProbe: [probeLibrary getProbeForVariable: "imitateProbVar"
				inClass: [self class]]];
  }

  [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdDist"
			      inClass: [self class]]];
  if(strcmp(habitImitateThresholdDist, "uniform") == 0) {
    [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdMin"
				inClass: [self class]]];
    [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdMax"
				inClass: [self class]]];
  }
  else {
    [pm addProbe:
	  [probeLibrary getProbeForVariable: "habitImitateThresholdMean"
			inClass: [self class]]];
    [pm addProbe: [probeLibrary getProbeForVariable: "habitImitateThresholdVar"
				inClass: [self class]]];
  }

  [pm addProbe: [probeLibrary getProbeForVariable: "memorySizeMin"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "memorySizeMax"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "strategySelectorFile"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "colour"
			      inClass: [self class]]];
  [pm addProbe: [probeLibrary getProbeForVariable: "probability"
			      inClass: [self class]]];
  [ObjectSaver save: self toFileNamed: filename withTemplate: pm];
  [pm drop];
  [strategies saveToFileNamed: strategySelectorFile];
  return self;
}

/*

writeHistory:

Write the parameters to the file

*/

-(void)writeHistory: (FILE *)fp {
  const char *nl;

  nl = [FearlusArguments nl];
  fprintf(fp, "Sub-population ID:\t%u%s", pin, nl);
  fprintf(fp, "Probability of land manager belonging to this sub-population:"
	  "\t%g%s", probability, nl);
  fprintf(fp, "Land Manager class:\t%s%s", landMgrClass, nl);
  fprintf(fp, "Strategy parameter ranges\tDistribution\tMinimum or Mean\t"
	  "Maximum or Variance%s"
	  "Neighbourhood weighting:\t%s\t%g\t%g%s"
	  "Probability of using imitative strategy when yield below habit-"
	  "imitation threshold:\t%s\t%g\t%g%s"
	  "Habit-imitation yield threshold:\t%s\t%g\t%g%s"
	  "Memory size:\tuniform\t%u\t%u%s",
	  nl, neighbourWeightDist,
	  ((strcmp(neighbourWeightDist, "uniform") == 0)
	   ? neighbourWeightMin : neighbourWeightMean),
	  ((strcmp(neighbourWeightDist, "uniform") == 0)
	   ? neighbourWeightMax : neighbourWeightVar), nl,
	  imitateProbDist,
	  ((strcmp(imitateProbDist, "uniform") == 0)
	   ? imitateProbMin : imitateProbMean),
	  ((strcmp(imitateProbDist, "uniform") == 0)
	   ? imitateProbMax : imitateProbVar), nl,
	  habitImitateThresholdDist,
	  ((strcmp(habitImitateThresholdDist, "uniform") == 0)
	   ? habitImitateThresholdMin : habitImitateThresholdMean),
	  ((strcmp(habitImitateThresholdDist, "uniform") == 0)
	   ? habitImitateThresholdMax : habitImitateThresholdVar), nl,
	  memorySizeMin, memorySizeMax, nl);
  fprintf(fp, "Strategy method selector%s", nl);
  [strategies writeHistory: fp];
  fprintf(fp, "End of parameters for sub-population ID:\t%u%s", pin, nl);
}

/*

clickMeToGetStrategySelectorGUI

Provide a method called from the probe map for this object that brings up a
GUI for entering/viewing the strategy selector data.

*/

-clickMeToGetStrategySelectorGUI {
  [strategies createGUIStuff];
  [strategies pack];
  [strategies configureWidths];
  return self;
}

/*

displayProbe

Provide a method to be called from the parameter probemap that brings up a
probe display for this object.

*/

-(void)displayProbe {
  CREATE_PROBE_DISPLAY(self);
}

/*

drop

Free all the stuff we've allocated

*/

-(void)drop {
  [[self getZone] free: label];
  [strategies drop];
  [managerList drop];
  [super drop];
}

/*

getManagerCount

Return the number of managers alive that belong to this sub-population
This method is most likely to be called from the ObserverSwarm.

*/

-(unsigned)getManagerCount {
  return [managerList getCount];
}


@end
